"""
OAuth2 Authorization Server for ActingWeb MCP clients.

This module implements ActingWeb as an OAuth2 authorization server that can issue
its own tokens to MCP clients while proxying user authentication to Google OAuth2.
"""
